// Ben Clewett, April 2004
//
// Very simple shell to execute SQL.  Gives a MySQL feel. Please use sqlite3 shell for any more than exeuting SQL.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/times.h>
#include <unistd.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <signal.h>
#include "sqlite3.h"
#include "lite_wrap.h"

void doLoop();
void showIntro();
void strip(char *s);
void execute_sql(char *s);
void showHelp();
void open_db(char *sName);
static void interrupt_handler(int);
void save_readline();
int localHandling(char *sql);
void showOptions();

char *sProgName;
sqlite3 *handle;
char sHistoryFile[256];

int main(int argc, char **argv)
{

	sProgName = argv[0];

	if (argc < 2) {
		showHelp();
		exit(1);
	}

	signal(SIGINT, interrupt_handler);

	open_db(argv[1]);

	showIntro();

	sprintf(sHistoryFile, "%s/.sw_shell", getenv("HOME"));
	read_history(sHistoryFile);

	doLoop();

	save_readline();

	return 0;

}

void doLoop()
{
	char sSQL[1024], *sLine = NULL, *sPrevLine = NULL;
	char sPrompt[20];

	sSQL[0] = 0;

	while (1) {

		if (!sSQL[0]) {
			sprintf(sPrompt, "\nsw_shell> ");
		} else {
			sprintf(sPrompt, "       -> ");
		}
		
		sLine = readline(sPrompt);
		strip(sLine);
		sprintf(sSQL, "%s%s%s", sSQL, (sSQL[0]) ? " " : "", sLine);

		if (sSQL[0] && (sSQL[strlen(sSQL) - 1] == ';' || sSQL[0] == '.')) {

			if (sPrevLine && strcmp(sSQL, sPrevLine)) add_history(sSQL);
			else if (!sPrevLine) add_history(sSQL);

			free(sPrevLine);
			sPrevLine = strdup(sSQL);

			if (!localHandling(sSQL)) execute_sql(sSQL);

			sSQL[0] = 0;
		}

	}



}

int localHandling(char *sql)
// Some sql handled locally
{
	char *sLine;
	int i;
	static char sNewSQL[1024];
	char sOption[256], sNull[10];

	if (sql[0] != '.') return 0;

	sNewSQL[0] = 0;

	sLine = strdup(sql);
	if (sql[strlen(sql) - 1] == ';') sql[strlen(sql) - 1] = 0;
	for (i = 0; sLine[i]; i++) sLine[i] = toupper(sLine[i]);

	sOption[0] = 0;
	sscanf(sql, "%s%s", sNull, sOption);

	if (!strncmp(sLine, ".EXIT", 5)) {
		interrupt_handler(0);
	}
	else if (!strncmp(sLine, ".SCHEMA", 7)) {
		if (sOption[0]) {
			sprintf(sNewSQL, "SELECT name, type, sql FROM sqlite_master WHERE sql NOT NULL AND type=='table' AND name = '%s'", sOption);
		} else {
			sprintf(sNewSQL, "SELECT name, type, sql FROM sqlite_master WHERE sql NOT NULL AND type=='table'");
		}
	} 
	else if (!strncmp(sLine, ".DATABASES", 7)) {
		strcpy(sNewSQL, "PRAGMA database_list");
	}
	else if (!strncmp(sLine, ".HELP", 5)) {
		showOptions();
	}
	else if (!strncmp(sLine, ".INDICES", 8)) {
		if (sOption[0]) {
			sprintf(sNewSQL, "SELECT name FROM sqlite_master WHERE type='index' AND tbl_name LIKE '%s' UNION ALL SELECT name FROM sqlite_temp_master WHERE type='index' AND tbl_name LIKE '%s' ORDER BY 1", sOption, sOption);
		} else {
			printf("Please supply table name.");
		}
	}
	else if (!strncmp(sLine, ".TABLES", 7)) {
		if (sOption[0]) {
			sprintf(sNewSQL, "SELECT name FROM sqlite_master WHERE type IN ('table','view') AND name LIKE '%%%s%%' UNION ALL SELECT name FROM sqlite_temp_master WHERE type IN ('table','view') AND name LIKE '%%%s%%' ORDER BY 1", sOption, sOption);
		} else {
			strcpy(sNewSQL, "SELECT name FROM sqlite_master WHERE type IN ('table','view') UNION ALL SELECT name FROM sqlite_temp_master WHERE type IN ('table','view') ORDER BY 1");
		}
	}
	else {
		printf("Unknown option.  ");
		showOptions();
	}

	if (sNewSQL[0]) execute_sql(sNewSQL);

	free (sLine);

	return 1;

}

void showOptions()
{
	printf("Options:\n");
	printf("\t.databases\n");
	printf("\t.exit\n");
	printf("\t.help\n");
	printf("\t.indices table\n");
	printf("\t.schema [table]\n");
	printf("\t.tables [%%like%%]\n");
}

void showIntro()
{
	printf("sw_shell.  Press CTRL-C to exit.  Enter .help for more options.\n\n");
}

void showHelp()
{
	printf("Usage:\n\t%s: database_name\n", sProgName);
	exit(0);
}

void strip(char *s)
{
	static int a, b;

	if (strlen(s) == 0) return;

	for (a = strlen(s) - 1; a >= 0 && s[a] < 33; a--) ;
	s[a + 1] = 0;

	if (a == -1) return;

	for (a = 0; a < strlen(s) - 1 && s[a] < 33; a++) ;
	for (b = a; a < strlen(s) ; a++) 
		s[a - b] = s[a];
	s[a - b] = 0;

	return;

}

void execute_sql(char *sql)
{
	lw_result *result;
	struct tms t0, t1;

	times(&t0);

	result = lw_query(handle, sql);

	if (result == NULL) {
		printf("Error: '%d', '%s'\n", lw_get_errno(), lw_get_errmsg());
		return;
	}

	if (result->row_count == 0) {
		printf("  (Empty set)");
	} else {
		lw_show_table(result, stdout);
	}

	lw_free_result(result);

	times(&t1);

	printf("  (%.3f seconds)\n", (t1.tms_utime - t0.tms_utime) / (double)sysconf(_SC_CLK_TCK));

}


void open_db(char *sName)
{
	int ret;

	if ((ret = sqlite3_open(sName, &handle)) != SQLITE_OK) {
		printf("Error opening: '%s': %d\n", sName, ret);
		exit(1);
	}

}

static void interrupt_handler(int i)
{
	save_readline();
	printf("\nBye!\n");
	exit(0);
}

void save_readline()
{
	write_history(sHistoryFile);
}
